"""Testing optpkg module
"""

import builtins
import sys
import types
from unittest import SkipTest, mock

import pytest
from packaging.version import Version

from nibabel.optpkg import optional_package
from nibabel.tripwire import TripWire, TripWireError


def assert_good(pkg_name, min_version=None):
    pkg, have_pkg, setup = optional_package(pkg_name, min_version=min_version)
    assert have_pkg
    assert sys.modules[pkg_name] == pkg
    assert setup() is None


def assert_bad(pkg_name, min_version=None):
    pkg, have_pkg, setup = optional_package(pkg_name, min_version=min_version)
    assert not have_pkg
    assert isinstance(pkg, TripWire)
    with pytest.raises(TripWireError):
        pkg.a_method
    with pytest.raises(SkipTest):
        setup()


def test_basic():
    # We always have os
    assert_good('os')
    # Subpackage
    assert_good('os.path')
    # We never have package _not_a_package
    assert_bad('_not_a_package')

    # Only disrupt imports for "nottriedbefore" package
    orig_import = builtins.__import__

    def raise_Exception(*args, **kwargs):
        if args[0] == 'nottriedbefore':
            raise Exception(
                'non ImportError could be thrown by some malfunctioning module '
                'upon import, and optional_package should catch it too'
            )
        return orig_import(*args, **kwargs)

    with mock.patch.object(builtins, '__import__', side_effect=raise_Exception):
        assert_bad('nottriedbefore')


def test_versions():
    fake_name = '_a_fake_package'
    fake_pkg = types.ModuleType(fake_name)
    assert 'fake_pkg' not in sys.modules
    # Not inserted yet
    assert_bad(fake_name)
    try:
        sys.modules[fake_name] = fake_pkg
        # No __version__ yet
        assert_good(fake_name)  # With no version check
        assert_bad(fake_name, '1.0')
        # We can make an arbitrary callable to check version
        assert_good(fake_name, lambda pkg: True)
        # Now add a version
        fake_pkg.__version__ = '2.0'
        # We have fake_pkg > 1.0
        for min_ver in (None, '1.0', Version('1.0'), lambda pkg: True):
            assert_good(fake_name, min_ver)
        # We never have fake_pkg > 100.0
        for min_ver in ('100.0', Version('100.0'), lambda pkg: False):
            assert_bad(fake_name, min_ver)
        # Check error string for bad version
        pkg, _, _ = optional_package(fake_name, min_version='3.0')
        try:
            pkg.some_method
        except TripWireError as err:
            assert str(err) == 'These functions need _a_fake_package version >= 3.0'
    finally:
        del sys.modules[fake_name]



"""Benchmarks for finite_range routine

Run benchmarks with::

    import nibabel as nib
    nib.bench()

Run this benchmark with::

    pytest -c <path>/benchmarks/pytest.benchmark.ini <path>/benchmarks/bench_finite_range.py
"""

import sys

import numpy as np
from numpy.testing import measure

from nibabel.volumeutils import finite_range  # NOQA

from .butils import print_git_title


def bench_finite_range():
    rng = np.random.RandomState(20111001)
    repeat = 10
    img_shape = (128, 128, 64, 10)
    arr = rng.normal(size=img_shape)
    sys.stdout.flush()
    print_git_title('\nFinite range')
    mtime = measure('finite_range(arr)', repeat)
    print('%30s %6.2f' % ('float64 all finite', mtime))
    arr[:, :, :, 1] = np.nan
    mtime = measure('finite_range(arr)', repeat)
    print('%30s %6.2f' % ('float64 many NaNs', mtime))
    arr[:, :, :, 1] = np.inf
    mtime = measure('finite_range(arr)', repeat)
    print('%30s %6.2f' % ('float64 many infs', mtime))
    # Int16 input, float output
    arr = np.random.random_integers(low=-1000, high=1000, size=img_shape)
    arr = arr.astype(np.int16)
    mtime = measure('finite_range(arr)', repeat)
    print('%30s %6.2f' % ('int16', mtime))
    sys.stdout.flush()



"""Module to help tests check script output

Provides class to be instantiated in tests that check scripts.  Usually works
something like this in a test module::

    from .scriptrunner import ScriptRunner
    runner = ScriptRunner()

Then, in the tests, something like::

    code, stdout, stderr = runner.run_command(['my-script', my_arg])
    assert_equal(code, 0)
    assert_equal(stdout, b'This script ran OK')
"""
import os
import sys
from os.path import dirname, isdir, isfile
from os.path import join as pjoin
from os.path import pathsep, realpath
from subprocess import PIPE, Popen

MY_PACKAGE = __package__


def local_script_dir(script_sdir):
    """Get local script directory if running in development dir, else None"""
    # Check for presence of scripts in development directory.  ``realpath``
    # allows for the situation where the development directory has been linked
    # into the path.
    package_path = dirname(__import__(MY_PACKAGE).__file__)
    above_us = realpath(pjoin(package_path, '..'))
    devel_script_dir = pjoin(above_us, script_sdir)
    if isfile(pjoin(above_us, 'setup.py')) and isdir(devel_script_dir):
        return devel_script_dir
    return None


def local_module_dir(module_name):
    """Get local module directory if running in development dir, else None"""
    mod = __import__(module_name)
    containing_path = dirname(dirname(realpath(mod.__file__)))
    if containing_path == realpath(os.getcwd()):
        return containing_path
    return None


class ScriptRunner:
    """Class to run scripts and return output

    Finds local scripts and local modules if running in the development
    directory, otherwise finds system scripts and modules.
    """

    def __init__(
        self,
        script_sdir='scripts',
        module_sdir=MY_PACKAGE,
        debug_print_var=None,
        output_processor=lambda x: x,
    ):
        """Init ScriptRunner instance

        Parameters
        ----------
        script_sdir : str, optional
            Name of subdirectory in top-level directory (directory containing
            setup.py), to find scripts in development tree.  Typically
            'scripts', but might be 'bin'.
        module_sdir : str, optional
            Name of subdirectory in top-level directory (directory containing
            setup.py), to find main package directory.
        debug_print_vsr : str, optional
            Name of environment variable that indicates whether to do debug
            printing or no.
        output_processor : callable
            Callable to run on the stdout, stderr outputs before returning
            them.  Use this to convert bytes to unicode, strip whitespace, etc.
        """
        self.local_script_dir = local_script_dir(script_sdir)
        self.local_module_dir = local_module_dir(module_sdir)
        if debug_print_var is None:
            debug_print_var = f'{module_sdir.upper()}_DEBUG_PRINT'
        self.debug_print = os.environ.get(debug_print_var, False)
        self.output_processor = output_processor

    def run_command(self, cmd, check_code=True):
        """Run command sequence `cmd` returning exit code, stdout, stderr

        Parameters
        ----------
        cmd : str or sequence
            string with command name or sequence of strings defining command
        check_code : {True, False}, optional
            If True, raise error for non-zero return code

        Returns
        -------
        returncode : int
            return code from execution of `cmd`
        stdout : bytes
            stdout from `cmd`
        stderr : bytes
            stderr from `cmd`
        """
        if isinstance(cmd, str):
            cmd = [cmd]
        else:
            cmd = list(cmd)
        if not self.local_script_dir is None:
            # Windows can't run script files without extensions natively so we need
            # to run local scripts (no extensions) via the Python interpreter.  On
            # Unix, we might have the wrong incantation for the Python interpreter
            # in the hash bang first line in the source file.  So, either way, run
            # the script through the Python interpreter
            cmd = [sys.executable, pjoin(self.local_script_dir, cmd[0])] + cmd[1:]
        if os.name == 'nt':
            # Quote any arguments with spaces. The quotes delimit the arguments
            # on Windows, and the arguments might be file paths with spaces.
            # On Unix the list elements are each separate arguments.
            cmd = [f'"{c}"' if ' ' in c else c for c in cmd]
        if self.debug_print:
            print(f"Running command '{cmd}'")
        env = os.environ
        if not self.local_module_dir is None:
            # module likely comes from the current working directory. We might need
            # that directory on the path if we're running the scripts from a
            # temporary directory
            env = env.copy()
            pypath = env.get('PYTHONPATH', None)
            if pypath is None:
                env['PYTHONPATH'] = self.local_module_dir
            else:
                env['PYTHONPATH'] = self.local_module_dir + pathsep + pypath
        proc = Popen(cmd, stdout=PIPE, stderr=PIPE, env=env)
        stdout, stderr = proc.communicate()
        if proc.poll() is None:
            proc.terminate()
        if check_code and proc.returncode != 0:
            raise RuntimeError(
                f"""Command "{cmd}" failed with
                stdout
                ------
                {stdout}
                stderr
                ------
                {stderr}
                """
            )
        opp = self.output_processor
        return proc.returncode, opp(stdout), opp(stderr)



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Testing mriutils module
"""


import pytest
from numpy.testing import assert_almost_equal

from ..mriutils import MRIError, calculate_dwell_time


def test_calculate_dwell_time():
    # Test dwell time calculation
    # This tests only that the calculation does what it appears to; needs some
    # external check
    assert_almost_equal(calculate_dwell_time(3.3, 2, 3), 3.3 / (42.576 * 3.4 * 3 * 3))
    # Echo train length of 1 is valid, but returns 0 dwell time
    assert_almost_equal(calculate_dwell_time(3.3, 1, 3), 0)
    with pytest.raises(MRIError):
        calculate_dwell_time(3.3, 0, 3.0)
    with pytest.raises(MRIError):
        calculate_dwell_time(3.3, 2, -0.1)



"""Generate mask and testing tractogram in known formats:

* mask: standard.nii.gz
* tractogram:

    * standard.trk
"""
import numpy as np

import nibabel as nib
from nibabel.streamlines import FORMATS
from nibabel.streamlines.header import Field


def mark_the_spot(mask):
    """Marks every nonzero voxel using streamlines to form a 3D 'X' inside.

    Generates streamlines forming a 3D 'X' inside every nonzero voxel.

    Parameters
    ----------
    mask : ndarray
        Mask containing the spots to be marked.

    Returns
    -------
    list of ndarrays
        All streamlines needed to mark every nonzero voxel in the `mask`.
    """

    def _gen_straight_streamline(start, end, steps=3):
        coords = []
        for s, e in zip(start, end):
            coords.append(np.linspace(s, e, steps))

        return np.array(coords).T

    # Generate a 3D 'X' template fitting inside the voxel centered at (0,0,0).
    X = [
        _gen_straight_streamline((-0.5, -0.5, -0.5), (0.5, 0.5, 0.5)),
        _gen_straight_streamline((-0.5, 0.5, -0.5), (0.5, -0.5, 0.5)),
        _gen_straight_streamline((-0.5, 0.5, 0.5), (0.5, -0.5, -0.5)),
        _gen_straight_streamline((-0.5, -0.5, 0.5), (0.5, 0.5, -0.5)),
    ]

    # Get the coordinates of voxels 'on' in the mask.
    coords = np.array(zip(*np.where(mask)))

    streamlines = [(line + c) * voxel_size for c in coords for line in X]

    return streamlines


if __name__ == '__main__':
    rng = np.random.RandomState(42)

    width = 4  # Coronal
    height = 5  # Sagittal
    depth = 7  # Axial

    voxel_size = np.array((1.0, 3.0, 2.0))

    # Generate a random mask with voxel order RAS+.
    mask = rng.rand(width, height, depth) > 0.8
    mask = (255 * mask).astype(np.uint8)

    # Build tractogram
    streamlines = mark_the_spot(mask)
    tractogram = nib.streamlines.Tractogram(streamlines)

    # Build header
    affine = np.eye(4)
    affine[range(3), range(3)] = voxel_size
    header = {
        Field.DIMENSIONS: (width, height, depth),
        Field.VOXEL_SIZES: voxel_size,
        Field.VOXEL_TO_RASMM: affine,
        Field.VOXEL_ORDER: 'RAS',
    }

    # Save the standard mask.
    nii = nib.Nifti1Image(mask, affine=affine)
    nib.save(nii, 'standard.nii.gz')

    # Save the standard tractogram in every available file format.
    for ext, cls in FORMATS.items():
        tfile = cls(tractogram, header)
        nib.streamlines.save(tfile, 'standard' + ext)



#!/usr/bin/env python
"""Script to auto-generate our API docs.
"""

import os
import re

# stdlib imports
import sys

# version comparison
from distutils.version import LooseVersion as V
from os.path import join as pjoin

# local imports
from apigen import ApiDocWriter

# *****************************************************************************


def abort(error):
    print(f'*WARNING* API documentation not generated: {error}')
    exit(1)


if __name__ == '__main__':
    package = sys.argv[1]
    outdir = sys.argv[2]
    try:
        other_defines = sys.argv[3]
    except IndexError:
        other_defines = True
    else:
        other_defines = other_defines in ('True', 'true', '1')

    # Check that the package is available. If not, the API documentation is not
    # (re)generated and existing API documentation sources will be used.

    try:
        __import__(package)
    except ImportError as e:
        abort('Can not import ' + package)

    module = sys.modules[package]

    # Check that the source version is equal to the installed
    # version. If the versions mismatch the API documentation sources
    # are not (re)generated. This avoids automatic generation of documentation
    # for older or newer versions if such versions are installed on the system.

    installed_version = V(module.__version__)

    version_file = pjoin('..', package, '_version.py')
    source_version = None
    if os.path.exists(version_file):
        # Versioneer
        from runpy import run_path

        try:
            source_version = run_path(version_file)['version']
        except (FileNotFoundError, KeyError):
            pass
        if source_version == '0+unknown':
            source_version = None
    if source_version is None:
        # Legacy fall-back
        info_file = pjoin('..', package, 'info.py')
        info_lines = open(info_file).readlines()
        source_version = '.'.join(
            [
                v.split('=')[1].strip(" '\n.")
                for v in info_lines
                if re.match('^_version_(major|minor|micro|extra)', v)
            ]
        )
    print('***', source_version)

    if source_version != installed_version:
        abort('Installed version does not match source version')

    docwriter = ApiDocWriter(package, rst_extension='.rst', other_defines=other_defines)
    docwriter.package_skip_patterns += [
        r'\.fixes$',
        r'\.fixes.*$',
        r'\.externals$',
        r'\.externals.*$',
        r'.*test.*$',
        r'\.info.*$',
        r'\.pkg_info.*$',
        r'\.py3k.*$',
        r'\._version.*$',
    ]
    docwriter.write_api_docs(outdir)
    docwriter.write_index(outdir, 'index', relative_to=outdir)
    print('%d files written' % len(docwriter.written_modules))



# init for sext package
"""Setuptools extensions

nibabel uses these routines, and houses them, and installs them.  nipy-proper
and dipy use them.
"""

import warnings

warnings.warn(
    """The nisext package is deprecated as of NiBabel 5.0 and will be fully
removed in NiBabel 6.0"""
)
